@@ -1,7 +1,7 @@
 
 /*
  * BZip3 - A spiritual successor to BZip2.
- * Copyright (C) 2022 Kamila Szewczyk
+ * Copyright (C) 2022-2023 Kamila Szewczyk
  *
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published by the Free
@@ -61,7 +61,7 @@ static const u32 crc32Table[256] = {
 };
 
 static u32 crc32sum(u32 crc, u8 * RESTRICT buf, size_t size) {
-    while (size--) crc = crc32Table[((crc >> 24) ^ *(buf++)) & 0xff] ^ (crc << 8);
+    while (size--) crc = crc32Table[((u8)crc ^ *(buf++)) & 0xff] ^ (crc >> 8);
     return crc;
 }
 
@@ -80,7 +80,7 @@ static u32 crc32sum(u32 crc, u8 * RESTRICT buf, size_t size) {
 
 #define MATCH 0xf2
 
-static u32 lzp_upcast(u8 * ptr) {
+static u32 lzp_upcast(const u8 * ptr) {
     // val = *(u32 *)ptr; - written this way to avoid UB
     u32 val;
     memcpy(&val, ptr, sizeof(val));
@@ -186,7 +186,7 @@ static s32 lzp_decode_block(const u8 * RESTRICT in, const u8 * in_end, s32 * RES
                 }
 
                 const u8 * ref = outs + val;
-                u8 * oe = out + len;
+                const u8 * oe = out + len;
                 if (oe > out_end) oe = out_end;
 
                 while (out < oe) *out++ = *ref++;
@@ -282,7 +282,7 @@ static int mrled(u8 * RESTRICT in, u8 * RESTRICT out, s32 outlen, s32 maxin) {
     while (op < outlen && ip < maxin) {
         c = in[ip++];
         if (t[c]) {
-            for (run = 0; (pc = in[ip++]) == 255 && ip < maxin; run += 255)
+            for (run = 0; ip < maxin && (pc = in[ip++]) == 255; run += 255)
                 ;
             run += pc + 1;
             for (; run > 0 && op < outlen; --run) out[op++] = c;
@@ -832,10 +832,11 @@ BZIP3_API int bz3_compress(u32 block_size, const u8 * const in, u8 * out, size_t
     *out_size += 13;
 
     // Compress and write the blocks.
+    size_t in_offset = 0;
     for (u32 i = 0; i < n_blocks; i++) {
         s32 size = block_size;
         if (i == n_blocks - 1) size = in_size % block_size;
-        memcpy(compression_buf, in, size);
+        memcpy(compression_buf, in + in_offset, size);
         s32 out_size_block = bz3_encode_block(state, compression_buf, size);
         if (bz3_last_error(state) != BZ3_OK) {
             s8 last_error = state->last_error;
@@ -847,6 +848,7 @@ BZIP3_API int bz3_compress(u32 block_size, const u8 * const in, u8 * out, size_t
         write_neutral_s32(out + *out_size, out_size_block);
         write_neutral_s32(out + *out_size + 4, size);
         *out_size += out_size_block + 8;
+        in_offset += size;
     }
 
     bz3_free(state);
